from .src.enigma import Enigma
from .src.plugboard import Plugboard
from .src.reflector import Reflector, REFLECTOR_A, REFLECTOR_B, REFLECTOR_C
from .src.rotor import Rotor, ROTOR_I, ROTOR_II, ROTOR_III, ROTOR_IV, ROTOR_V


UPPERCASE_LETTERS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

class Rotor:
    """
    Represents a Rotor that maps characters pairwise and turns the rotor to the right notches.
    """

    def __init__(self, wiring:str=None, name:str=None, ring_position:int=1, rotor_position:int=1, notch_position:str=None):
        if wiring is not None:
            self.wiring = wiring
        else:
            self.wiring = UPPERCASE_LETTERS
        self.name = name
        self.notch_position = notch_position.upper()
        self.ring_position = ring_position
        self.rotor_position = rotor_position
        
        self.rev_wiring = ["0"] * 26
        for i in range(0, len(self.wiring)):
            index = ord(self.wiring[i])- ord("A")
            self.rev_wiring[index] = chr(ord("A") + i)
        self.rev_wiring = ''.join(self.rev_wiring)
    
    def __setattr__(self, name: str, value) -> None:
        self.__dict__[name] = value


    def __eq__(self, rotor) -> bool:
        return self.name == rotor.name

    def __str__(self) -> str:
        return self.name    
    
    def __len__(self):
        return len(self.name)

    def set_ring_position(self, ring_position:int = 1):
        """
        Convenience method for setting the ring position af a rotor.
        """
        assert type(ring_position) is int
        assert ring_position < 27
        self.ring_position = ring_position

    def encipher(self, key: str, mapping: str) -> str:
        """
        Maps a character to its counterpart and the defined mapping.        
        """
        assert type(key) is str
        assert len(key) == 1
        key = key.upper()
        assert key in UPPERCASE_LETTERS
        index = (ord(key) - ord("A"))%26
        rotor_shift = self.rotor_position - self.ring_position
        index = (index + rotor_shift)%26
        letter = mapping[index]
        letter = chr(ord("A") + (ord(letter) - ord("A") + 26 - rotor_shift) % 26)

        return letter

    def encipher_forward(self, key: str) -> str:
        """
        Simulates the forward pass of a character through a rotor.
        """
        return self.encipher(key=key, mapping=self.wiring)

    def encipher_backwards(self, key: str) -> str:
        """
        Simulates the backward pass of a character through a rotor.
        """
        return self.encipher(key=key, mapping=self.rev_wiring)

    def notch(self) -> None:
        """
        Turns a rotor by one position.
        """
        self.rotor_position = self.rotor_position + 1 if self.rotor_position <26 else 1

    def is_in_turnover_pos(self) -> bool:
        """
        Determines if a rotor is in turnover position for the next rotor.
        """
        return self.rotor_position == ord(self.notch_position) - ord("A")


ROTOR_I = Rotor(wiring="EKMFLGDQVZNTOWYHXUSPAIBRCJ", name="I", ring_position=1, rotor_position=1, notch_position='R')
ROTOR_II = Rotor(wiring="AJDKSIRUXBLHWTMCQGZNPYFVOE", name="II", ring_position=1, rotor_position=1, notch_position='F')
ROTOR_III = Rotor(wiring="BDFHJLCPRTXVZNYEIWGAKMUSQO", name="III", ring_position=1, rotor_position=1, notch_position='W')
ROTOR_IV = Rotor(wiring="ESOVPZJAYQUIRHXLNFTGKDCMWB", name="IV", ring_position=1, rotor_position=1, notch_position='K')
ROTOR_V = Rotor(wiring="VZBRGITYUPSDNHLXAWMJQOFECK", name="V", ring_position=1, rotor_position=1, notch_position='A')


UPPERCASE_LETTERS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

class Plugboard:
    """
    Represents a Plugboard that maps characters according to the plug settings of the enigma machine
    """

    def __init__(self, plug_settings: str = "AA BB CC DD EE FF GG HH II JJ") -> None:
        self.plug_input = plug_settings
        self.plug_settings = [(pair[0].upper(), pair[1].upper()) for pair in plug_settings.split(' ')]

        assert len(self.plug_settings)<=10

        #check for duplicate values in settings
        for mapping_pair in self.plug_settings:
            temp=self.plug_settings.copy()
            temp.remove(mapping_pair)
            assert mapping_pair[0] not in str(temp)
            assert mapping_pair[1] not in str(temp)

        self.mapping = {letter:letter for letter in UPPERCASE_LETTERS}

        for pair0, pair1 in self.plug_settings: 
            self.mapping.update({pair0:pair1})
            self.mapping.update({pair1:pair0})

    def map_plug(self, key:str):
        """
        Maps a character to its counterpart according to a plug setting.
        """
        return self.mapping[key]
    
    def __str__(self):
        """
        Pretty display.
        """
        return str(self.plug_input)
    





import pytest
from ..src.rotor import ROTOR_I, ROTOR_II, ROTOR_III, ROTOR_IV, ROTOR_V
from ..src.reflector import REFLECTOR_A, REFLECTOR_B, REFLECTOR_C
from ..src.enigma import Enigma
from ..src.plugboard import Plugboard

@pytest.fixture
def EnigmaMachineRotorRingPositions():
    EnigmaMachineRotorRingPositions = Enigma(reflector=REFLECTOR_B, 
                        left_rotor=ROTOR_III, 
                        middle_rotor=ROTOR_II, 
                        right_rotor=ROTOR_I, 
                        rotor_positions="16 10 1" ,
                        ring_positions="3 1 17",
                        plugboard=Plugboard()
                        )
    return EnigmaMachineRotorRingPositions

@pytest.fixture
def EnigmaMachineRotorRingPlugs():
    EnigmaMachineRotorRingPlugs = Enigma(reflector=REFLECTOR_B, 
                        left_rotor=ROTOR_III, 
                        middle_rotor=ROTOR_II, 
                        right_rotor=ROTOR_I, 
                        rotor_positions="16 10 1" ,
                        ring_positions="3 1 17",
                        plugboard=Plugboard("AQ JH FR EW TZ")
                        )
    return EnigmaMachineRotorRingPlugs


def test_encrypt_phrase(EnigmaMachineRotorRingPositions):
    encrypted_text=EnigmaMachineRotorRingPositions.encipher('THISBETTERBEWORKINGFORREALTHISTIME')
    assert encrypted_text == 'vjrqfaanthvttssrrgujemcrfmhxqmsgtf'.upper()


def test_encrypt_phrase2(EnigmaMachineRotorRingPlugs):
    encrypted_text=EnigmaMachineRotorRingPlugs.encipher('THISBETTERBEWORKINGFORREALTHISTIME')
    assert encrypted_text == 'mjfaroueydvnqsjffgubwziqymiiamxgzz'.upper()





from ..src.plugboard import *
import pytest


def test_plugboard_correct_mapping():
    PLUGBOARD = Plugboard(plug_settings="AB FG DE")
    assert PLUGBOARD.map_plug('A')=='B'
    assert PLUGBOARD.map_plug('B')=='A'
    assert PLUGBOARD.map_plug('F')=='G'
    assert PLUGBOARD.map_plug('G')=='F'
    assert PLUGBOARD.map_plug('Z')=='Z'
    
def test_plugboard_no_settings():
    PLUGBOARD = Plugboard()
    assert PLUGBOARD.map_plug('A')=='A'
    assert PLUGBOARD.map_plug('B')=='B'

def test_plugboard_duplicate_settings():
    with pytest.raises(AssertionError):
        PLUGBOARD = Plugboard(plug_settings="AB AE")

def test_plugboard_too_many_plug_settings():
    with pytest.raises(AssertionError):
        PLUGBOARD = Plugboard(plug_settings="AB CD EF GH IJ KL MN OP QR ST UV WX")


from ..src.rotor import *
import pytest


def test_ROTORI_encipher_forward():
    ROTOR_I.rotor_position = 1
    assert ROTOR_I.encipher_forward('B')=='K'

def test_ROTORI_encipher_backwards():
    ROTOR_I.rotor_position = 1
    assert ROTOR_I.encipher_backwards('K')=='B' 

def test_ROTORI_encipher_int():
    with pytest.raises(AssertionError):
        ROTOR_I.encipher_forward(9)

def test_ROTORI_encipher_two_str():
    with pytest.raises(AssertionError):
        ROTOR_I.encipher_forward('AA')

def test_ROTORI_encipher_back_int():
    with pytest.raises(AssertionError):
        ROTOR_I.encipher_backwards(9)

def test_ROTORI_encipher_back_two_str():
    with pytest.raises(AssertionError):
        ROTOR_I.encipher_backwards('AA')

def test_ROTORI_notch():
    ROTOR_I.rotor_position = 1
    ROTOR_I.notch()
    assert ROTOR_I.rotor_position == 2
    ROTOR_I.notch()
    assert ROTOR_I.rotor_position == 3

def test_ROTORI_notch_over():
    ROTOR_I.rotor_position = 25
    ROTOR_I.notch()
    assert ROTOR_I.rotor_position == 26
    ROTOR_I.notch()
    assert ROTOR_I.rotor_position == 1
    ROTOR_I.notch()
    assert ROTOR_I.rotor_position == 2


def test_ROTORI_ringsetting_encipher_forward():
    ROTOR_I.rotor_position = 1
    ROTOR_I.set_ring_position(2)
    assert ROTOR_I.encipher_forward('K')=='A'
    ROTOR_I.rotor_position = 1
    ROTOR_I.set_ring_position(26)
    assert ROTOR_I.encipher_forward('Z')=='D'


def test_ROTORI_is_at_notch():
    ROTOR_I.rotor_position = ord(ROTOR_I.notch_position)-ord("A")
    assert ROTOR_I.is_in_turnover_pos()